/*
  This script creates stored package sp_purge for purging data in the Nolio database.
  Procedure purge_execution_jobs purges offline execution jobs; purge_audit purges audit
  records.

  ==========================
  purge_execution_jobs
  ==========================
  Purge execution jobs and its associated records one day at a time starting from the oldest
  records.  The following tables are being purged:
    offline_distribution_events
    offline_exec_param_links
    offline_exec_servers
    offline_execution_jobs
    offline_flow_events
    offline_manual_operations
    offline_parameter_requests
    offline_parameters
    offline_propagation_events
    offline_step_events
    offline_steps
    offline_user_events

  Parameters:
  retention_days       Number of days to retain execution job records.  Default is 90 days
  time_limit           Number of seconds after which no command is executed.  The procedure
                       will stop when the time limit is reached and the last command
                       (started before the time limit) is complete.  The default is no time 
                       limit (null).
  time_delay           Number of seconds to delay between each delete command.  The default
                       is null, no delay.*
  delete_chunk_size    Maximum number of rows to delete per command.  The default is 10000
                       rows.

  Example:
  The following example purges jobs older than 120 days, with a time limit so that no command is
  executed after 3600 seconds (1 hour)

  begin
    sp_purge.purge_execution_jobs(retention_days=>120, time_limit=>3600);
  end;


  ==========================
  sp_purge_audit
  ==========================
  Purge audit records
  The following tables are affected by the purging script: 
    auditingentry 
    revision_auditingentry 
    All table with _aud suffix 
        The following tables are excluded: 
          health_record_aud 
          health_round_samples_aud 
          health_sample_aud 
          health_threshold_aud
          published_flow_aud

  Parameters:
  time_limit         Number of seconds after which no command is executed.  The procedure
                     will stop when the time limit is reached and the last command
                     (started before the time limit) is complete.  The default is no time 
                     limit (null).
  time_delay         Number of seconds to delay between each delete command.  The default
                     is null, no delay.*
  delete_chunk_size  Maximum number of rows to delete per command.  The default is 10000
                     rows.

  Example:
  The following example purges audit records with a time limit so that no command is
  executed after 3600 seconds (1 hour)

  begin
    sp_purge.purge_audit(time_limit=>3600);
  end;


  *When time_delay parameter is specified, the procedures uses DBMS_LOCK.SLEEP to wait between
  commands.  If time_delay is specified, the Nolio user must have EXECUTE privilege on
  SYS.DBMS_LOCK package.  If time_delay is not specified or it is NULL, then the privilege is
  not required.  Execute the following statement from a DBA user to grant the privilege.
    grant execute on SYS.DBMS_LOCK to <nolio owner>;

*/

-- create an index on offline_execution_jobs.end_time and id if it doesn't exists
DECLARE
  is_exist int;
  sql_stmt varchar2(200);
BEGIN
  select count(1) into is_exist 
    from user_tables t join user_tab_columns c1 
              on t.table_name = c1.table_name
              and c1.column_name = 'END_TIME'
          join user_ind_columns ic1
              on ic1.table_name = c1.table_name
              and ic1.column_name = c1.column_name
              and ic1.column_position = 1
          join user_tab_columns c2
              on t.table_name = c2.table_name
              and c2.column_name = 'ID'
          join user_ind_columns ic2
              on ic2.table_name = c2.table_name
              and ic2.column_name = c2.column_name
              and ic2.index_name = ic1.index_name
              and ic2.column_position = 2
    where t.table_name = 'OFFLINE_EXECUTION_JOBS';
    
  IF is_exist = 0 THEN
    sql_stmt := 'CREATE INDEX IX_OFFLINE_EXEC_JOB_ENDTIME_ID ON OFFLINE_EXECUTION_JOBS(END_TIME, ID)';
    dbms_output.put_line(sql_stmt);  
    EXECUTE IMMEDIATE sql_stmt;
  END IF;
END;
/



-- create global temporary table GTT_PURGE_AUDIT and GTT_PURGE_AUDIT2 to temporary store audit ID/REV for deletion 
DECLARE 
  IS_EXIST INT;
  SQL_STMT VARCHAR2(4000);
BEGIN
  SELECT COUNT(1) INTO IS_EXIST FROM USER_TABLES WHERE TABLE_NAME = 'GTT_PURGE_AUDIT';
  
  IF IS_EXIST = 0 THEN
    SQL_STMT := 'CREATE GLOBAL TEMPORARY TABLE GTT_PURGE_AUDIT'
               ||' (ID NUMBER(19,0),'
               ||' PRIMARY KEY(ID))'
               ||' ON COMMIT PRESERVE ROWS';
                
    DBMS_OUTPUT.PUT_LINE(SQL_STMT);  
    EXECUTE IMMEDIATE sql_stmt;
  END IF;

  SELECT COUNT(1) INTO IS_EXIST FROM USER_TABLES WHERE TABLE_NAME = 'GTT_PURGE_AUDIT2';
  
  IF IS_EXIST = 0 THEN
    SQL_STMT := 'CREATE GLOBAL TEMPORARY TABLE GTT_PURGE_AUDIT2'
               ||' (ID NUMBER(19,0),'
               ||' ID2 NUMBER(19,0),'
               ||' PRIMARY KEY(ID, ID2))'
               ||' ON COMMIT PRESERVE ROWS';
                
    DBMS_OUTPUT.PUT_LINE(SQL_STMT);  
    EXECUTE IMMEDIATE sql_stmt;
  END IF;
END;
/

CREATE OR REPLACE PACKAGE sp_purge AS
  PROCEDURE purge_execution_jobs (retention_days int := 90,
                                  time_limit int := null,
                                  time_delay int := null,
                                  delete_chunk_size int := 10000);

  PROCEDURE purge_audit (time_limit int := null,
                         time_delay int := null,
                         delete_chunk_size int := 10000);
END sp_purge;
/

CREATE OR REPLACE PACKAGE BODY sp_purge AS
  PROCEDURE delete_chunk (sql_part varchar2, 
                          delete_chunk_size int,
                          time_delay int,
                          p1 timestamp)
  IS
    sql_stmt varchar2(4000);
  BEGIN
    sql_stmt :=   'BEGIN'
               ||'  LOOP'
               ||case when time_delay is null then '' else '    DBMS_LOCK.SLEEP('||time_delay||');' end
               ||'    '||sql_part||' and rownum <= '||delete_chunk_size||';'
               ||'    EXIT WHEN SQL%ROWCOUNT < '||delete_chunk_size||';'
               ||'    COMMIT;'
               ||'  END LOOP;'
               ||'  COMMIT;'
               ||'END;';

    EXECUTE IMMEDIATE sql_stmt USING p1;       
  END delete_chunk;                      

  PROCEDURE purge_execution_jobs (retention_days int := 90,
                                  time_limit int := null,
                                  time_delay int := null,
                                  delete_chunk_size int := 10000)
  IS
    cur_date timestamp;
    cutoff_date timestamp := sysdate - retention_days;
    stop_time date := null;
    sql_stmt varchar2(4000);
  
  BEGIN
    IF time_limit IS NOT NULL THEN
      stop_time := sysdate + time_limit/24.0/60.0/60.0;
    END IF;
    
    select trunc(min(end_time)) + 1 into cur_date from offline_execution_jobs where end_time is not null;
  
    WHILE cur_date < cutoff_date AND (stop_time is null OR sysdate < stop_time) LOOP
  
      dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss')||': deleting offline execution jobs older than '||cur_date||' ...');
  
      delete_chunk ('DELETE FROM offline_distribution_events WHERE id IN (SELECT /*+ORDERED USE_NL(e)*/ e.id FROM offline_execution_jobs j, offline_flow_events e WHERE j.id = e.job_id AND j.end_time < :1 )', delete_chunk_size, time_delay, cur_date);
  
      delete_chunk ('delete from offline_exec_param_links where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_exec_servers where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_flow_events where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_manual_operations where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_parameter_requests where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_parameters where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_propagation_events where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_step_events where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_steps where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
      delete_chunk ('delete from offline_user_events where job_id in (select id from offline_execution_jobs where end_time < :1)', delete_chunk_size, time_delay, cur_date);
  
      delete_chunk ('delete from offline_execution_jobs where end_time < :1', delete_chunk_size, time_delay, cur_date);
  
      cur_date := cur_date + 1;
    END LOOP;
  
    IF (sysdate >= stop_time) THEN
      dbms_output.put_line('Exceeded Time limit.  Exiting procedure.');
    ELSE
      dbms_output.put_line('Purge done.');
    END IF;
  END purge_execution_jobs;

  PROCEDURE purge_audit (time_limit int := null,
                         time_delay int := null,
                         delete_chunk_size int := 10000)
  IS
    TYPE varchar2_table IS TABLE OF VARCHAR2(4000) INDEX BY BINARY_INTEGER;

    delete_stmts varchar2_table;
    i BINARY_INTEGER := 0;
    new_line varchar2(1) := chr(10);
    stop_time date := null;
    sql_stmt varchar2(4000);
    delay_stmt varchar2(300) := '';
    prev_parent_table varchar2(30) := ' ';
    parent_table varchar2(30);
    child_table varchar2(30);
    last_id number(19) := -1; 
    rev_index varchar2(30);
      
    -- get _aud tables based on FK relationship
    -- The following tables should not be purged and are excluded from the query:
    --   published_flow_aud, health_record_aud, health_round_samples_aud, health_threshold_aud, health_sample_aud
    CURSOR cur_aud_table IS 
    select t2.parent_table, t2.child_table, max(ic.index_name) parent_table_rev_index
    from (select substr(path, 2, instr(path, '/', 1, 2) - 2) parent_table,
                 child_table,
                 level_
          from (select SYS_CONNECT_BY_PATH (t.table_name, '/') || '/' path,
                       level level_,
                       t.table_name child_table
                from user_tables t 
                     join user_tab_columns c
                       on c.table_name = t.table_name and c.column_name = 'REV'
                     left join (select rk.table_name parent_table, fk.table_name child_table
                                from user_constraints fk, user_constraints rk
                                where rk.constraint_name = fk.r_constraint_name
                                and fk.constraint_type = 'R'
                                and rk.table_name like '%\_AUD' escape '\') fk2
                      on fk2.child_table = t.table_name
                where t.table_name like '%\_AUD' escape '\' 
                and t.table_name not in ('PUBLISHED_FLOW_AUD', 'HEALTH_RECORD_AUD', 'HEALTH_ROUND_SAMPLES_AUD', 'HEALTH_THRESHOLD_AUD', 'HEALTH_SAMPLE_AUD')
                start with fk2.parent_table is null
                connect by prior t.table_name = fk2.parent_table
               ) t1
          ) t2
          left join user_ind_columns ic
           on t2.parent_table = ic.table_name and ic.column_name = 'REV' and column_position = 1
    group by t2.parent_table, t2.level_, t2.child_table
    order by t2.parent_table, t2.level_ desc, t2.child_table;

  BEGIN
    IF time_limit IS NOT NULL THEN
      stop_time := sysdate + time_limit/24.0/60.0/60.0;
    END IF;  

    IF time_delay IS NOT NULL THEN
      delay_stmt := new_line||'    DBMS_LOCK.SLEEP('||time_delay||');';
    END IF;  

    dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss')||': Deleting audit entries...');
  
    OPEN cur_aud_table;
    LOOP
      FETCH cur_aud_table INTO parent_table, child_table, rev_index;
      EXIT WHEN cur_aud_table%NOTFOUND;
      
      IF prev_parent_table <> parent_table and parent_table = child_table THEN
        
        -- table has no child.  Loop and delete a chunk at a time until time limit is reached
        sql_stmt := 'BEGIN'
                    ||new_line
                    ||'  WHILE (:stop_time is null OR sysdate < :stop_time) LOOP'
                    ||delay_stmt
                    ||new_line
                    ||'    delete from'
                    ||'    (select'
                    ||case when rev_index is null then '' else ' /*+LEADING(gtt aud) USE_NL(aud) INDEX(aud '||rev_index||')*/' end
                    ||'     1 from '||parent_table||' aud, GTT_PURGE_AUDIT gtt'
                    ||'     where gtt.id = aud.rev)'
                    ||'    where rownum <= '||delete_chunk_size||';'
                    ||new_line
                    ||'    EXIT WHEN SQL%ROWCOUNT < '||delete_chunk_size||';'
                    ||new_line
                    ||'    COMMIT;'
                    ||new_line
                    ||'  END LOOP;'
                    ||new_line
                    ||'  COMMIT;'
                    ||new_line
                    ||'END;';
                    
        i := i + 1;
        delete_stmts(i) := sql_stmt;

      ELSE
        -- aud table has children.  Get rows from the aud tables and delete them
        -- from aud table and its children
        IF prev_parent_table <> parent_table THEN

          sql_stmt := 'DECLARE'
                      ||new_line
                      ||'  row_count NUMBER(12) := '||delete_chunk_size||';'
                      ||new_line
                      ||'BEGIN'
                      ||new_line
                      ||'  WHILE ((:stop_time is null OR sysdate < :stop_time) AND row_count = '||delete_chunk_size||') LOOP'
                      ||delay_stmt
                      ||new_line
                      ||'    EXECUTE IMMEDIATE ''truncate table GTT_PURGE_AUDIT2'';'
                      ||new_line
                      ||'    insert into GTT_PURGE_AUDIT2 (id, id2)'
                      ||'    select'
                      ||case when rev_index is null then '' else ' /*+LEADING(gtt aud) USE_NL(aud) INDEX(aud '||rev_index||')*/' end
                      ||'    aud.ID, aud.REV'
                      ||'    from GTT_PURGE_AUDIT gtt, '||parent_table||' aud'
                      ||'    where aud.rev = gtt.id'
                      ||'    and rownum <= '||delete_chunk_size||';'
                      ||new_line
                      ||'    row_count := SQL%ROWCOUNT;'
                      ||new_line
                      ||'    EXIT WHEN row_count = 0;';
        END IF;
  
        sql_stmt := sql_stmt 
                  ||delay_stmt
                  ||new_line
                  ||'    delete from'
                  ||'    (select /*+LEADING(gtt aud) USE_NL(aud)*/ 1'
                  ||'     from '||child_table||' aud, GTT_PURGE_AUDIT2 gtt'
                  ||'     where gtt.id2 = aud.rev and gtt.id = aud.id);'
                  ||new_line
                  ||'    COMMIT;';
  
        IF parent_table = child_table THEN
          sql_stmt := sql_stmt
                      ||new_line
                      ||'  END LOOP;'
                      ||new_line
                      ||'END;';

          i := i + 1;
          delete_stmts(i) := sql_stmt;
          
        END IF;
      END IF;  
      
      prev_parent_table := parent_table;
    END LOOP;
    CLOSE cur_aud_table;

    -- add statement to delete from revision_auditingentry and auditingentry
    sql_stmt := 'DECLARE'
                ||new_line
                ||'  row_count NUMBER(12) := '||delete_chunk_size||';'
                ||new_line
                ||'BEGIN'
                ||new_line
                ||'  WHILE ((:stop_time is null OR sysdate < :stop_time) AND row_count = '||delete_chunk_size||') LOOP'
                ||delay_stmt
                ||new_line
                ||'    EXECUTE IMMEDIATE ''truncate table GTT_PURGE_AUDIT2'';'
                ||new_line
             
                -- get ID's from revision_auditingentry and store in GTT_PURGE_AUDIT2
                ||'    insert into GTT_PURGE_AUDIT2 (id, id2)'
                ||'    select  /*+LEADING(gtt aud) USE_NL(aud)*/ aud.designrevisionentity_id, aud.auditEntries_id'
                ||'    from GTT_PURGE_AUDIT gtt, revision_auditingentry aud'
                ||'    where aud.designrevisionentity_id = gtt.id'
                ||'    and rownum <= '||delete_chunk_size||';'

                 -- break if no row found
                ||new_line
                ||'    row_count := SQL%ROWCOUNT;'
                ||new_line
                ||'    EXIT WHEN row_count = 0;'

                 -- delete revision_auditingentry
                ||delay_stmt
                ||new_line
                ||'    delete from'
                ||'    (select /*+LEADING(gtt aud) USE_NL(aud)*/ 1'
                ||'     from revision_auditingentry aud, GTT_PURGE_AUDIT2 gtt'
                ||'     where gtt.id2 = aud.auditEntries_id and gtt.id = aud.designrevisionentity_id);'
                ||new_line
                ||'    COMMIT;'

                 -- delete auditingentry
                ||delay_stmt
                ||new_line
                ||'    delete from'
                ||'    (select /*+LEADING(gtt aud) USE_NL(aud)*/ 1'
                ||'     from auditingentry aud, GTT_PURGE_AUDIT2 gtt'
                ||'     where gtt.id2 = aud.id);'
                ||new_line
                ||'    COMMIT;'
                
                ||new_line
                ||'  END LOOP;'
                ||new_line
                ||'END;';

    i := i + 1;
    delete_stmts(i) := sql_stmt;

  
    -- loop to get 100,000 IDs at a time from designrevisionentity and delete the corrresponding REV rows
    -- from each aud table.
    WHILE (stop_time is null OR sysdate < stop_time) LOOP
      
      dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss')||': Purging REV > '||last_id||'...');
    
      IF time_delay IS NOT NULL THEN
        EXECUTE IMMEDIATE 'BEGIN '||delay_stmt||' END;';
      END IF;  
    
      EXECUTE IMMEDIATE 'truncate table GTT_PURGE_AUDIT';

      insert into GTT_PURGE_AUDIT
      select id from (select  d.id from designrevisionentity d    
                      where d.processStatus is not null
                      and d.id > last_id
                      order by 1)
      where rownum <= 100000;
      
      EXIT WHEN SQL%ROWCOUNT = 0;
  
      -- loop to delete from each aud table until time limit is reached
      i := delete_stmts.FIRST;
      WHILE i IS NOT NULL AND (stop_time is null OR sysdate < stop_time) LOOP
      
        /*
        if (last_id = -1) then
          dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss'));
          dbms_output.put_line(delete_stmts(i));
        end if;
        */
        
        EXECUTE IMMEDIATE delete_stmts(i) USING stop_time;
        i := delete_stmts.NEXT(i);
      END LOOP;    
      
      select max(id) into last_id from GTT_PURGE_AUDIT;
    END LOOP;    


    IF sysdate >= stop_time THEN
      dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss')||'  Exceeded time limit.  Exiting procedure.');
    ELSE
      dbms_output.put_line(to_char(sysdate, 'yyyy/mm/dd hh24:mi:ss')||'  Purge done.');
    END IF;
  END purge_audit;

END sp_purge;
/

